# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Dict
from .dao import Tools, ToolConf, AbstractRequest
from .model.crop import SegmentationRequest
from .model.convert import ConvertRequest, Option
from .model import (
    AlignmentRequest,
    DeconvolutionRequest, 
    EdofRequest,
    YoloSegdetectRequest
)

def _create_template_requests() -> Dict | AbstractRequest:

    # TODO Do we want to have proper concrete types for these requests?
    # We could autogenerate them from the Java classes as we did for the
    # typescript JIT client.

    # We could also move the existing dao concrete types in all the python
    # tools to dao here then import them into each tool.
    # TODO Submit a ticket to do that.

    ret = {}

    # Crop
    request = SegmentationRequest() # Default values are set in the dataclass
    ret[Tools.CROP] = request

    # Deconvolution
    # We set some defaults here for users to start with.
    request = DeconvolutionRequest()
    request.inputPath = None
    request.stain1 = [0.563, 0.72, 0.406]
    request.stain2 = [0.216, 0.801, 0.558]
    request.stain1Max = 2.0
    request.stain2Max = 1.0
    request.alpha = 1
    request.beta = 0.15
    request.intensityNorm = 240
    request.grayscale = False
    request.stain1ImageOutput = None
    request.stain2ImageOutput = None
    request.normalizedImageOutput = None
    request.imageStackOutput = None
    
    ret[Tools.DECONVOLUTION] = request

    # TODO Use concrete python class?
    request = YoloSegdetectRequest()
    ret[Tools.YOLO_SEGDETECT] = request

    # TODO Use concrete python class?
    request = AlignmentRequest()
    ret[Tools.ALIGNMENT] = request

    request = EdofRequest(inputFiles=[],outputPath=None)
    request.gradientKernel= 5
    request.imageNoiseFilter = 3
    request.zmapNoiseFilter = 3
    request.lowPass = 2
    request.threads = 2
    request.dryRun = False
    request.outputZmap = None
    request.inputZmap = None
    ret[Tools.EDOF] = request

    request = ConvertRequest()
    request.options = [Option.overwrite]
        #  You may use this string to override the reset of the parameters
        #  and provide all the command line apart from the input image and the
        #  output image as a string. Setting this will negative the other
        #  options. This is useful if you want to provide a command line
        #  which is already known. It is not intended to be part of the web UI
        #  but is useful for the CLI.
    request.cmd = None
    request.uploadDir = None  # StorageKey for the dir to which we will upload the converted file(s)
    request.useGeoJson = True
        # If any of the following patterns are present in out_file, they will
        # be replaced with the indicated metadata value from the input file.
        #
        #    Pattern:    Metadata value:
        #    ---------------------------
        #    %s        series index
        #    %n        series name
        #    %c        channel index
        #    %w        channel name
        #    %z        Z index
        #    %t        T index
        #    %A        acquisition timestamp
        #    %x        row index of the tile
        #    %y        column index of the tile
        #    %m        overall tile index
        #
        # For example: "tile_%x_%y.tiff" for the name of each tiled tiff extracted.
        # If you do not set outfile correctly then BioFormats will not generate the
        # correct images in the directory.
    request.outFile = None
        # This is the value of the [-compression codec] option in BioFormats.
        # Set to compress to a specific format.
    request.compression = None
        # The series which we want to extract for conversion from the
        # input file. Usually used (None) which defaults to (0) to extract the first series.
    request.series = None
    request.map= None
        # [-range start end] option on BioFormats command line.
    request.range = None
        # Set the rectangle to crop from the image.
        # [-crop x,y,w,h] example: request["crop"] = {100,100,400,400}
        # x,y top left corner, w width, h height
    request.crop = None
        # a polygon array of all the regions to crop and use for the converted
        # image example:
        # request["manualRegions"] = [ {
        #         "npoints" : 4,
        #         "xpoints" : [ 0, 0, 100, 100 ],
        #         "ypoints" : [ 0, 100, 100, 0 ]
        #       }, {
        #         "npoints" : 4,
        #         "xpoints" : [ 100, 100, 200, 200 ],
        #         "ypoints" : [ 0, 100, 100, 0 ]
        #       }, {
        #         "npoints" : 4,
        #         "xpoints" : [ 200, 200, 300, 300 ],
        #         "ypoints" : [ 0, 100, 100, 0 ]
        #       } ]
    request.manualRegions = None
        # [-channel channel] option on BioFormats command line.
    request.channel = None
        # [-z Z]
    request.z = None
        # [-timepoint timepoint]
    request.timepoint = None
        # [-option key value]
    request.option = None
        # [-tilex tileSizeX]
    request.tilex = None
        # [-tiley tileSizeY]
    request.tiley = None
        # [-pyramid-scale scale]
    request.pyramidScale = None
        # [-swap dimensionsOrderString]
    request.swap = None
        # [-fill color] 0-255
    request.fill = None
        # [-pyramid-resolutions numResolutionLevels]
    request.pyramidResolutions =  None
    
    ret[Tools.CONVERT] = request

    return ret


template_request_dict: Dict = _create_template_requests()


def create_template_request(type: ToolConf):
    if type not in template_request_dict:
        raise ValueError(f"No template request for type {type.name}")
    return template_request_dict[type]



# generated by datamodel-codegen:
#   filename:  ConvertRequest-schema.json
#   timestamp: 2025-11-24T20:11:47+00:00

# Procedure:
# 1. Generate from schema:
#  datamodel-codegen \
#--input ../cimg-api/src/main/resources/schema/ConvertRequest-schema.json  
#--input-file-type jsonschema 
#--output src/orgjax/pyjit/model/convert.py 
#--output-model-type dataclasses.dataclass 
#
# 2. Remove classes such as StorageKey and Protocol, instead import them from dao.py
# 3. Change the main request to inherit from AbstractRequest
# 4. Make sure that the defaults from ../templates.py are the same.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from dataclasses import field
from ..dao import StorageKey, AbstractRequest
from . import Info


class Option(Enum):
    debug = 'debug'
    stich = 'stich'
    separate = 'separate'
    merge = 'merge'
    expand = 'expand'
    bigtiff = 'bigtiff'
    nobigtiff = 'nobigtiff'
    cache = 'cache'
    nogroup = 'nogroup'
    nolookup = 'nolookup'
    autoscale = 'autoscale'
    version = 'version'
    noupgrade = 'noupgrade'
    padded = 'padded'
    nosas = 'nosas'
    novalid = 'novalid'
    validate = 'validate'
    precompressed = 'precompressed'
    overwrite = 'overwrite'
    showinf = 'showinf'


class Compression(Enum):
    Uncompressed = 'Uncompressed'
    JPEG = 'JPEG'
    JPEG2000 = 'JPEG2000'
    Lossy = 'Lossy'
    LZW = 'LZW'
    zlib = 'zlib'

@dataclass
class Point:
    x: float
    y: float


@dataclass
class Dimension:
    width: float
    height: float


@dataclass
class ConvertRequest(AbstractRequest):
    
    info: Optional[Info] = field(default_factory=lambda: Info.CONVERT)

    useGeoJson: Optional[bool] = False
    info: Optional[Info] = None
    options: Optional[List[Option]] = None
    cmd: Optional[str] = None
    inFile: Optional[str] = None
    uploadDir: Optional[StorageKey] = None
    outFile: Optional[str] = None
    compression: Optional[Compression] = None
    series: Optional[int] = None
    cacheDir: Optional[str] = None
    map: Optional[str] = None
    range: Optional[List[int]] = None
    crop: Optional[Rectangle] = None
    manualRegions: Optional[List[Polygon]] = None
    channel: Optional[int] = None
    z: Optional[int] = None
    timepoint: Optional[int] = None
    option: Optional[List[str]] = None
    tilex: Optional[int] = None
    tiley: Optional[int] = None
    pyramidScale: Optional[int] = None
    swap: Optional[str] = None
    fill: Optional[int] = None
    pyramidResolutions: Optional[int] = None


@dataclass
class Rectangle:
    x: float
    y: float
    width: float
    height: float
    empty: bool
    minX: float
    minY: float
    maxX: float
    maxY: float
    centerX: float
    centerY: float
    bounds2D: Optional[Rectangle2D] = None
    location: Optional[Point] = None
    size: Optional[Dimension] = None


@dataclass
class Rectangle2D:
    minX: float
    minY: float
    maxX: float
    maxY: float
    centerX: float
    centerY: float
    x: float
    y: float
    width: float
    height: float
    empty: bool
    bounds2D: Optional[Rectangle2D] = None
    bounds: Optional[Rectangle] = None


@dataclass
class Polygon:
    npoints: int
    xpoints: Optional[List[int]] = None
    ypoints: Optional[List[int]] = None
    bounds: Optional[Rectangle] = None
    boundingBox: Optional[Rectangle] = None
    bounds2D: Optional[Rectangle2D] = None



from enum import Enum
from dataclasses import dataclass, field
from typing import List, Optional
from ..dao import AbstractRequest, StorageKey

class Info(Enum):
    COMPOUND = "COMPOUND"
    CROP = "CROP"
    FIND = "FIND"
    HTQUANT = "HTQUANT"
    ALIGNMENT = "ALIGNMENT"
    EDOF = "EDOF"
    KIDNEY_CLASSIFIER = "KIDNEY_CLASSIFIER"
    DECONVOLUTION = "DECONVOLUTION"
    YOLO_SEGDETECT = "YOLO_SEGDETECT"
    RETINAL_LAYER = "RETINAL_LAYER"
    STAIN_ADAPTER = "STAIN_ADAPTER"
    UNKNOWN = "UNKNOWN"
    CONVERT = "CONVERT"


@dataclass
class AlignmentRequest(AbstractRequest):

    info: Optional[Info] = field(default_factory=lambda: Info.ALIGNMENT)

    # Inputs
    align: Optional[StorageKey] = None
    reference: Optional[StorageKey] = None
    shg: Optional[StorageKey] = None

    # Outputs
    siftOutput: Optional[StorageKey] = None
    overlayedResult: Optional[StorageKey] = None

    # New parameters
    transform: Optional[bool] = True
    flip: Optional[bool] = True
    downsample: Optional[bool] = False
    max_size: Optional[int] = 9e7
    gamma: Optional[float] = 1.0

    resizedN: Optional[StorageKey] = None
    resizedSF: Optional[StorageKey] = None

    # Test parameters local to this repo,
    # not in the json used on the server.
    test: Optional[bool] = False

@dataclass(kw_only=True)
class EdofRequest(AbstractRequest):
    
    info: Optional[Info] = field(default_factory=lambda: Info.EDOF)
    
    # Inputs
    inputFiles: List[StorageKey]
    outputPath: StorageKey

    inputZmap: Optional[StorageKey] = None

    # Size of kernel used for computation of image gradients.Must be 1, 3, 5, 7.
    gradientKernel: Optional[int] = 5
    # Size of median filter used to reduce noise in the z-map. Must be 0, 3 or 5
    imageNoiseFilter: Optional[int] = 3
    # Size of median filter used to reduce noise in the z-map. Must be 0, 3 or 5
    zmapNoiseFilter: Optional[int] = 3
    # zmap lowpass filter object size.
    lowPass: Optional[float] = 2.0

    # Test parameters local to this repo,
    # not in the json used on the server.
    test: Optional[bool] = False


@dataclass(kw_only=True)
class DeconvolutionRequest(AbstractRequest):
    
    info: Optional[Info] = field(default_factory=lambda: Info.DECONVOLUTION)
    
    # Inputs
    inputPath: Optional[StorageKey] = None
    stain1: str | tuple[float, float, float] = "hematoxylin"
    stain2: str | tuple[float, float, float] = "eosin"
    stain1Max: float = 2.0
    stain2Max: float = 1.0
    alpha: int = 1
    beta: float = 0.15
    intensityNorm: int = 240
    grayscale: bool = False
    # Outputs
    stain1ImageOutput: Optional[StorageKey] = None
    stain2ImageOutput: Optional[StorageKey] = None
    normalizedImageOutput: Optional[StorageKey] = None
    imageStackOutput: Optional[StorageKey] = None

    # Test parameters local to this repo,
    # not in the json used on the server.
    test: Optional[bool] = False
    
@dataclass(kw_only=True)
class YoloSegdetectRequest(AbstractRequest):
    
    info: Optional[Info] = field(default_factory=lambda: Info.YOLO_SEGDETECT)

    # Inputs
    inputPaths: Optional[List[StorageKey]] = None
    outputPath: Optional[StorageKey] = None
    weightsPath: Optional[str] = None
    downsamplingFactor: Optional[int] = 5
    visualize: Optional[bool] = False
    overlapX: Optional[int] = 0
    overlapY: Optional[int] = 0
    confidence: Optional[float] = 0.6
    iouThreshold: Optional[float] = 0.5
    nmsThreshold: Optional[float] = 0.3
    saveSegment: Optional[bool] = False
    level: Optional[int] = 2

    # Set if we want the activity to store the
    # classification to claris to be run.
    db_upsert: Optional[bool] = False

    # Test parameters local to this repo,
    # not in the json used on the server.
    test: Optional[bool] = False




# generated by datamodel-codegen:
#   filename:  SegmentationRequest-schema.json
#   timestamp: 2025-11-20T19:48:15+00:00

# Procedure:
# 1. Generate from schema:
#  datamodel-codegen \
# --input ../cimg-api/src/main/resources/schema/SegmentationRequest-schema.json \
# --input-file-type jsonschema \
# --output src/orgjax/pyjit/model/crop.py \
# --output-model-type dataclasses.dataclass
#
# 2. Remove classes such as StorageKey and Protocol, instead import them from dao.py
# 3. Change the main request to inherit from AbstractRequest
# 4. Make sure that the defaults from ../templates.py are the same.


from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from ..dao import StorageKey, AbstractRequest
from . import Info


@dataclass
class ParticleSettings:
    min: float = 1200000
    max: float = 2147483647
    cMin: float = 0.1
    cMax: float = 1.0
    hardBounds: bool = True


@dataclass
class ImagePreprocessSettings:
    rollingBallRadius: int = 50
    sigma: float = 100
    accuracy: float = 0.002
    autoThreshold: Optional[str] = "Otsu"


class Order(Enum):
    NONE = "NONE"
    ROWS = "ROWS"
    COLUMS = "COLUMS"


@dataclass
class Point:
    x: float
    y: float


@dataclass
class Dimension:
    width: float
    height: float


class Type(Enum):
    FeatureCollection = "FeatureCollection"


class Type1(Enum):
    name = "name"
    link = "link"


@dataclass
class Object:
    pass


class Type2(Enum):
    Feature = "Feature"


class Type3(Enum):
    Polygon = "Polygon"


class Type4(Enum):
    MultiPolygon = "MultiPolygon"


class Type5(Enum):
    Point = "Point"


class Type6(Enum):
    LineString = "LineString"


class Type7(Enum):
    MultiLineString = "MultiLineString"


class Type8(Enum):
    GeometryCollection = "GeometryCollection"


class Order1(Enum):
    IMAGE = "IMAGE"
    GRAPH = "GRAPH"


@dataclass
class TileConfiguration:
    clip: bool = True
    alwaysTile: bool= False
    expandCroppedFiles: bool = True
    tileSize: Optional[List[int]] = field(default_factory=lambda: [4096, 4096])
    order: Optional[Order1] = Order1.IMAGE
    cropFillColor: Optional[List[int]] = None


@dataclass
class Crs:
    type: Optional[Type1] = None
    properties: Optional[Dict[str, Object]] = None


@dataclass
class Polygon1:
    type: Type3
    crs: Optional[Crs] = None
    bbox: Optional[List[float]] = None
    coordinates: Optional[List[List]] = None


@dataclass
class MultiPolygon:
    type: Type4
    crs: Optional[Crs] = None
    bbox: Optional[List[float]] = None
    coordinates: Optional[List[List[List]]] = None


@dataclass
class Point1:
    type: Type5
    crs: Optional[Crs] = None
    bbox: Optional[List[float]] = None
    coordinates: Optional[Any] = None


@dataclass
class LineString:
    type: Type6
    crs: Optional[Crs] = None
    bbox: Optional[List[float]] = None
    coordinates: Optional[List] = None


@dataclass
class MultiLineString:
    type: Type7
    crs: Optional[Crs] = None
    bbox: Optional[List[float]] = None
    coordinates: Optional[List[List]] = None


MultiPoint = LineString


@dataclass
class RegionSettings:
    minimumAspectRatio: float = 0.15
    margin: int = 10
    findRegions: bool = True
    showRegions: bool = False
    manualRegionResolution: int = 0
    manualRegions: Optional[List[Polygon]] = field(default_factory=lambda: [])
    userRegions: Optional[FeatureCollection] = None
    generateProofImage: bool = True
    generatePreviews: bool = True
    previewSize: int = 2048
    zip: bool = True
    saveAsJson: bool = True
    fontSize: int= 160
    cropFoundRegionsAsRectangles: bool= True
    forceSquare: bool= False
    cropAsGreyScale: bool= False
    reference: Optional[Rectangle] = None
    order: Optional[Order] = Order.ROWS


@dataclass
class Rectangle:
    x: float
    y: float
    width: float
    height: float
    empty: bool
    minX: float
    minY: float
    maxX: float
    maxY: float
    centerX: float
    centerY: float
    bounds2D: Optional[Rectangle2D] = None
    location: Optional[Point] = None
    size: Optional[Dimension] = None


@dataclass
class Rectangle2D:
    x: float
    y: float
    width: float
    height: float
    minX: float
    minY: float
    maxX: float
    maxY: float
    centerX: float
    centerY: float
    empty: bool
    bounds2D: Optional[Rectangle2D] = None
    bounds: Optional[Rectangle] = None


@dataclass
class Polygon:
    npoints: int
    xpoints: Optional[List[int]] = None
    ypoints: Optional[List[int]] = None
    bounds: Optional[Rectangle] = None
    bounds2D: Optional[Rectangle2D] = None
    boundingBox: Optional[Rectangle] = None


@dataclass
class FeatureCollection:
    type: Type
    crs: Optional[Crs] = None
    bbox: Optional[List[float]] = None
    features: Optional[List[Feature]] = None


@dataclass
class Feature:
    type: Type2
    crs: Optional[Crs] = None
    bbox: Optional[List[float]] = None
    properties: Optional[Dict[str, Object]] = None
    geometry: Optional[
        Union[
            Feature,
            Polygon1,
            MultiPolygon,
            FeatureCollection,
            Point1,
            MultiPoint,
            MultiLineString,
            LineString,
            GeometryCollection,
        ]
    ] = None
    id: Optional[str] = None


@dataclass
class GeometryCollection:
    type: Type8
    crs: Optional[Crs] = None
    bbox: Optional[List[float]] = None
    geometries: Optional[
        List[
            Union[
                Feature,
                Polygon1,
                MultiPolygon,
                FeatureCollection,
                Point1,
                MultiPoint,
                MultiLineString,
                LineString,
                GeometryCollection,
            ]
        ]
    ] = None

@dataclass
class SegmentationRequest(AbstractRequest):
    
    info: Optional[Info] = field(default_factory=lambda: Info.CROP)
    
    bytesPerPixel: int = 0
    findResolution: int = -2147483648
    optimumFindSize: int = 16000000
    cropResolution: int = 0
    saveRequest: bool= True
    imageJPlugin: bool = False
    image: Optional[str] = None
    outputDir: Optional[StorageKey] = None
    cropIndices: Optional[List[int]] = None
    particleSettings: Optional[ParticleSettings] = field(default_factory=ParticleSettings)
    imagePreprocess: Optional[ImagePreprocessSettings] = field(default_factory=ImagePreprocessSettings)
    regionSettings: Optional[RegionSettings] = field(default_factory=RegionSettings)
    prefix: Optional[str] = ""
    tileConfiguration: Optional[TileConfiguration] = field(default_factory=TileConfiguration)
    version: Optional[str] = None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Copyright [2025] The Jackson Laboratory


from dataclasses import dataclass, field, asdict, is_dataclass

from typing import List, Optional, Any
from pathlib import Path
from enum import Enum
from urllib.parse import urlparse

import os

from . import shared_config


def convert(obj):
     if is_dataclass(obj):
         return {k: convert(v) for k, v in asdict(obj).items()}
     elif isinstance(obj, Enum):
         return obj.value
     elif isinstance(obj, list):
         return [convert(i) for i in obj]
     elif isinstance(obj, dict):
         return {k: convert(v) for k, v in obj.items()}
     else:
         return obj

class State:
    SUBMITTED: str = "SUBMITTED"
    RUNNING: str = "RUNNING"
    COMPLETE: str = "COMPLETE"
    ERROR: str = "ERROR"
    NONE: str = "NONE"
    CANCELLED: str = "CANCELLED"


@dataclass
class UntypedStatus:
    """
    Based on cimg-api/org.jax.cimg.api.evt.UntypedStatus
    which we use in Image Tools.
    These fields are shared across multiple tools.
    Do not change them.
    """

    message: Optional[str] = None
    state: str = State.NONE
    complete: float = 0
    errorStack: Optional[str] = None
    workflowId: Optional[str] = None
    results: List[str] = field(default_factory=lambda: [""])
    submissionInput: Optional[str] = None
    userName: Optional[str] = None


class Protocol(str, Enum):
    GS = "GS"
    S3 = "S3"
    NIO = "NIO"


@dataclass
class StorageKey:

    bucket: str
    object: str
    protocol: Optional[str] = None  # One of Protocol
    endpoint: Optional[str] = None
    test: Optional[bool] = False

    def to_dict(self):
        # This is require to make
        # json serialization to work.
        return {
            "bucket": self.bucket,
            "object": self.object,
            "protocol": self.protocol,
            "endpoint": self.endpoint,
        }

    def clone(self) -> "StorageKey":
        # You can trick python here using a string to let
        # it return a type but you cannot return a StorageKey
        # because we have not defined it yet.
        # Now wash your eyes.
        return StorageKey(
            bucket=self.bucket,
            object=self.object,
            protocol=self.protocol,
            endpoint=self.endpoint,
            test=self.test,
        )

    def set_file_name(self, file_name: str) -> "StorageKey":
        """
        Replace the file name in the object path with the given file_name.
        """
        p = Path(self.object)
        if p.parent == Path("."):
            self.object = file_name
        else:
            if p.is_dir() or self.object.endswith("/"):
                self.object = str(p / file_name)
            else:
                self.object = str(p.parent / file_name)
        self.object = self.object.replace("\\", "/")

        return self

    def to_uri(self, create=False) -> str:
        if self._is_local():
            file_path: str = None
            if self.endpoint is None:
                file_path: str = "{}/{}".format(self.bucket, self.object)
            else:
                file_path: str = "{}/{}/{}".format(
                    self.endpoint, self.bucket, self.object
                )

            NIO_STORAGE_ROOT: str = shared_config.NIO_STORAGE_ROOT
            if NIO_STORAGE_ROOT:
                file_path = "{}/{}".format(NIO_STORAGE_ROOT, file_path)

            # Remove double slashes
            file_path = file_path.replace("//", "/")
            file_path = file_path.replace("\\", "/")

            if create:
                path: Path = Path(file_path)
                os.makedirs(path.parent, exist_ok=True)
                path.touch(exist_ok=True)
            return file_path
        else:
            return f"{self.protocol.lower()}://{self.bucket}/{self.object}"

    def _is_local(self) -> bool:
        if self.protocol == Protocol.NIO:
            return True
        if self.test:
            return True

        USE_NIO_STORAGE: bool = shared_config.USE_NIO_STORAGE
        if USE_NIO_STORAGE:
            return True
        return False

    @classmethod
    def from_uri(cls, uri: str) -> "StorageKey":
        """
        Parse a URI and return a StorageKey instance.
        Supports formats like:
          - gs://bucket/object
          - s3://bucket/object
          - nio://bucket/object
          - /local/path/to/object (NIO/local)
        """
        parsed = urlparse(uri)
        if parsed.scheme in ("gs", "s3", "nio"):
            protocol = parsed.scheme.upper()
            bucket = parsed.netloc
            object_ = parsed.path.lstrip("/")
            object_ = object_.replace("\\", "/")
            return cls(bucket=bucket, object=object_, protocol=protocol)
        else:
            # Assume local path (NIO)
            # You may want to split the path differently depending on your conventions
            parts = Path(uri).parts
            if len(parts) >= 2:
                bucket = parts[0]
                object_ = str(Path(*parts[1:]))
            else:
                bucket = ""
                object_ = uri
            object_ = object_.replace("\\", "/")
            return cls(bucket=bucket, object=object_, protocol=Protocol.NIO)


@dataclass
class AbstractRequest:
    info: Optional[Any] = None

    def __eq__(self, other):
        if not isinstance(other, AbstractRequest):
            return False
        return self.info == other.info

    def __hash__(self):
        return hash(self.info)

    def to_dict(self):
         return convert(self)
     
@dataclass
class Input:
    """
    Input is the single class we send to any workflow in JIT.
    request is the actual request object specific to the workflow.
    """

    image: StorageKey
    request: AbstractRequest
    result: Optional[Any] = None
    userName: Optional[str] = None
    # Tried using field(default=None, metadata=config(field_name="userName")) to make is user_name
    # but did not work.


class ToolConf:
    """
    Class containing queue name constants.
    """

    def __init__(
        self,
        name: Optional[str],
        queue: Optional[str],
        workflow_name: Optional[str],
        workflow_query: Optional[str],
        # SharedConfig.TEMPORAL_NAMESPACE defaults to "default"
        namespace: Optional[str] = os.getenv("TEMPORAL_NAMESPACE", "cimg-dev"),
    ):

        self.name = name
        self.queue = queue
        self.workflow_name = workflow_name
        self.workflow_query = workflow_query
        self.namespace = namespace


class Tools:

    # Not a ProcessType value
    NONE: ToolConf = ToolConf("NONE", "NONE", None, None)

    # -------------------------
    # NOTE These strings must match org.jax.cimg.api.dao.ProcessType
    # precisely.
    # -------------------------
    CROP: ToolConf = ToolConf(
        "CROP", "SEGMENTATION_TASK_QUEUE", "Segmentation Workflow", "Segmentation Query"
    )
    FIND: ToolConf = ToolConf("FIND", "FIND_TASK_QUEUE", "Find Workflow", "Find Query")
    HTQUANT: ToolConf = ToolConf(
        "HTQUANT", "HTQUANT_QUEUE", "HtQuant Workflow", "HtQuant Query"
    )
    ALIGNMENT: ToolConf = ToolConf(
        "ALIGNMENT", "SIFT_QUEUE", "Sift Workflow", "Sift Query"
    )
    EDOF: ToolConf = ToolConf("EDOF", "EDOF_QUEUE", "EDOF Workflow", "EDOF Query")
    KIDNEY_CLASSIFIER: ToolConf = ToolConf(
        "KIDNEY_CLASSIFIER",
        "orchestration-task",
        "OrchestrationWorkflow",
        "classifierStatus",
    )
    DECONVOLUTION: ToolConf = ToolConf(
        "DECONVOLUTION",
        "DECONVOLUTION_QUEUE",
        "Color Deconvolution Workflow",
        "Color Deconvolution Query",
    )
    YOLO_SEGDETECT: ToolConf = ToolConf(
        "YOLO_SEGDETECT",
        "YOLO_SEGDETECT_QUEUE",
        "YoloSegdetect Workflow",
        "YoloSegdetect Query",
    )
    RETINAL_LAYER: ToolConf = ToolConf(
        "RETINAL_LAYER",
        "RETINAL_LAYER_QUEUE",
        "Retinal Layer Workflow",
        "Retinal Layer Query",
    )
    STAIN_ADAPTER: ToolConf = ToolConf(
        "STAIN_ADAPTER",
        "STAIN_ADAPTER_QUEUE",
        "Stain Adapter Workflow",
        "Stain Adapter Query",
    )
    CONVERT: ToolConf = ToolConf(
        "CONVERT", "CONVERT_TASK_QUEUE", "Convert Workflow", "Convert Query"
    )


def blank_key() -> StorageKey:
    return StorageKey("", "", test=True)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Copyright [2025] The Jackson Laboratory

from typing import Optional, Any, Dict, List
from .dao import UntypedStatus, Input, ToolConf, StorageKey
from .auth import get_access_token

from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
import base64


import sys
import os
import time
import requests
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

import random
import string
from pathlib import Path
import json
from time import sleep


def run_workflow(
    input: Input,
    tool: ToolConf,
    server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
    query_interval=2,
    subscriber=None,
) -> UntypedStatus:
    """
    Run a workflow given an input and tool configuration.
    """

    # Submit the workflow
    monitor_uri = submit_workflow(input, tool, server_uri, query_interval)

    # Monitor until complete
    return monitor_workflow(monitor_uri, query_interval)


def monitor_workflow(
    monitor_uri: str, query_interval=2, subscriber=None
) -> UntypedStatus:
    """
    Monitor a workflow until completion.
    """

    status: UntypedStatus = None

    # Monitor workflow status
    while not is_final(status):
        try:
            token = get_access_token()
            headers = {
                "Authorization": f"Bearer {token}",
                "Content-Type": "application/json",
            }
            response = requests.get(monitor_uri, headers=headers)
            response.raise_for_status()
            status: UntypedStatus = response.json()

            if subscriber is not None:
                subscriber(status)

            complete: float = float(status["complete"])
            complete_percent: float = complete * 100.0
            state: str = status["state"]
            message: str = status["message"]
            sys.stdout.write(f"\r{state} {complete_percent}% {message}")
            sys.stdout.flush()
            if is_final(status):
                break

            time.sleep(query_interval)

        except Exception as e:
            print(f"Exception occurred while monitoring workflow: {e}")
            break

    return status


def submit_workflow(
    input: Input,
    tool: ToolConf,
    server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
    query_interval=2,
) -> Any:
    """
    Submit a workflow given an input and tool configuration.
    """

    instr = create_instruction(input)
    token = get_access_token()

    s = requests.Session()
    retries = Retry(total=3, backoff_factor=1, status_forcelist=[502, 503, 504])
    s.mount("https://", HTTPAdapter(max_retries=retries))
    
    url = f"{server_uri}/process"
    params = {"type": tool.name}
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
    
    with s.post(url, params=params, json=instr, timeout=(5, 120), headers=headers) as response:
        response.raise_for_status()  # Raises an error for bad responses
        monitor: str = response.json()  # Or response.text if you expect plain text
        monitor_uri = monitor["response"]
        print(f"Monitor URI: {monitor_uri}")
        sleep(2)  # Give server a moment to start processing, not strictly necessary
        # but gives a better user experience.
        return monitor_uri


def delete(
    file_or_dir: StorageKey,
    server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
    type: str = "File",
) -> None:
    """
    Delete a file or directory.
    """

    blob: Dict = create_blob(file_or_dir, type)
    token: str = get_access_token()

    url = f"{server_uri}/delete"
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}

    with requests.post(url, json=blob, headers=headers) as response:
        response.raise_for_status()  # Raises an error for bad responses


def download(
    file: StorageKey,
    server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
    dir: Path = Path("./tmp"),
) -> Path:
    """
    Delete a file or directory.
    """

    blob: Dict = create_blob(file)
    token: str = get_access_token()

    url = f"{server_uri}/download"
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
    os.makedirs(dir, exist_ok=True)
    filename = os.path.basename(file.object)
    output_path = os.path.join(dir, filename)

    params = {"info": json.dumps(blob)}
    with requests.get(url, params=params, headers=headers, stream=True) as response:
        response.raise_for_status()
        with open(output_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
    return Path(output_path)


def get_report(uri: str, download_dir="./tmp") -> List[str]:
    image: StorageKey = StorageKey.from_uri(
        uri
    )  # The first result is the text file with all the paths.
    local_path: Path = Path(download_dir)  # Specify the dir where we want it.
    loc: Path = download(image, dir=local_path)

    with open(loc, "r", encoding="utf-8") as f:
        lines = f.readlines()
        return lines


def upload(
    file: Path,
    dir: StorageKey,
    server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
) -> Path:
    """
    Delete a file or directory.
    """

    token: str = get_access_token()

    url = f"{server_uri}/upload"
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
        "uri": dir.to_uri(),
        "key": get_upload_key(file),
    }

    with open(file, "rb") as f:
        files = {"file": (file.name, f)}
        with requests.post(url, headers=headers, files=files) as response:
            response.raise_for_status()

    return file


UPLOAD_KEY = os.getenv("UPLOAD_KEY", "A secret key base64 encoded==")


def get_upload_key(file_path: Path) -> str:
    key = get_key()

    # AES in ECB mode (since Java Cipher.getInstance("AES") defaults to AES/ECB/PKCS5Padding)
    cipher = Cipher(algorithms.AES(key), modes.ECB(), backend=default_backend())
    encryptor = cipher.encryptor()

    # Prepare plaintext (filename as UTF-8)
    plaintext = file_path.name.encode("utf-8")

    # Apply PKCS7 padding (block size = 128 bits for AES)
    padder = padding.PKCS7(128).padder()
    padded = padder.update(plaintext) + padder.finalize()

    # Encrypt
    encrypted_bytes = encryptor.update(padded) + encryptor.finalize()

    # Return base64 string
    return base64.b64encode(encrypted_bytes).decode("utf-8")


def get_key():
    # Decode the base64 key into raw bytes
    key_bytes = base64.b64decode(UPLOAD_KEY)
    return key_bytes


def create_instruction(input: Input) -> Dict:
    
    req = input.request
    if not isinstance(req, dict):
        req = req.to_dict()
    instr = {
        "file": create_blob(input.image),
        "request": req,
    }
    return instr


def create_blob(storage_key, type: str = "File") -> Dict:
    blob = {
        "name": os.path.basename(storage_key.object),
        "project": {
            "bucket": storage_key.bucket,
            "name": "Python api submission to {}".format(storage_key.bucket),
        },
        "type": type,
        "relPath": storage_key.object,
    }
    return blob


def is_final(status: UntypedStatus) -> bool:
    if status is None:
        return False
    state: str = status["state"]
    return state in ("COMPLETE", "ERROR", "CANCELLED")


def is_storage_result(result: str) -> bool:

    if "://" in result:
        return True
    return False


def random_string(size: Optional[int] = 4) -> str:
    # Generate a n-character random string (uppercase + lowercase letters)
    rand_str = "".join(random.choice(string.ascii_letters) for _ in range(size))
    return rand_str
